#include <alloc.h>
#include "windows.h"
#include "define.h"
#include "flexappl.h"
#include "flexnet.h"
#include "agw.h"
#include "tools.h"
#include "log.h"
#include "buffers.h"

extern QSO AGW_qso[MAX_USERSTREAM + 1];

static char FLEXNET_myCall[10];
static int  FLEXNET_nQsoNum[MAX_USERSTREAM + 1];
static FLEXNET_EVENT* FLEXNET_pEventHead = NULL;
static CRITICAL_SECTION cs;

int  FLEXNET_initialized = FALSE;

/* WDXNFLX.DLL declarations */
HINSTANCE gLibWdxnFlx = NULL;
typedef int (*WDXNFLX_AXINIT)();
WDXNFLX_AXINIT WDXNFLX_axinit;
typedef u16 (*WDXNFLX_L2GETSABM)(byte far*);
WDXNFLX_L2GETSABM WDXNFLX_l2getsabm;
typedef int (*WDXNFLX_L2STATE)(int);
WDXNFLX_L2STATE WDXNFLX_l2state;
typedef void (*WDXNFLX_L2CLRI)(int);
WDXNFLX_L2CLRI WDXNFLX_l2clri;
typedef void (*WDXNFLX_L2IACK)(int);
WDXNFLX_L2IACK WDXNFLX_l2iack;
typedef void (*WDXNFLX_L2SENDCHAR)(u16, byte);
WDXNFLX_L2SENDCHAR WDXNFLX_l2sendchar;
typedef void (*WDXNFLX_L2PACK)(int);
WDXNFLX_L2PACK WDXNFLX_l2pack;
typedef const FRAME* (*WDXNFLX_L2GETF)(int);
WDXNFLX_L2GETF WDXNFLX_l2getf;
typedef i8 (*WDXNFLX_L2IALLOC)(int, int);
WDXNFLX_L2IALLOC WDXNFLX_l2ialloc;
typedef u16 (*WDXNFLX_L2UNACK)(int);
WDXNFLX_L2UNACK WDXNFLX_l2unack;
typedef u16 (*WDXNFLX_L2CONNECT)(const byte *, const byte *);
WDXNFLX_L2CONNECT WDXNFLX_l2connect;
typedef const INFO* (*WDXNFLX_L2GETI)(int);
WDXNFLX_L2GETI WDXNFLX_l2geti;
typedef byte* (*WDXNFLX_STRCALL)(byte*, const byte*);
WDXNFLX_STRCALL WDXNFLX_strcall;
typedef void (*WDXNFLX_L2STOPQSO)(int);
WDXNFLX_L2STOPQSO WDXNFLX_l2stopqso;
typedef void (*WDXNFLX_L2SABMRESP)(int);
WDXNFLX_L2SABMRESP WDXNFLX_l2sabmresp;
typedef void (*WDXNFLX_L2SABMDM)(int);
WDXNFLX_L2SABMDM WDXNFLX_l2sabmdm;


/* --------------------------------------------------------------------------------------------- 
   ------ Add a new event in the event list ---------------------------------------------------- 
   --------------------------------------------------------------------------------------------- 
   IMPORTANT : take care entering a critical section before calling that function */
static void FLEXNET_addEvent(int nStreamNum, int nEventID, const unsigned char* pszData, int nDataLen)
{
	FLEXNET_EVENT* pEvent;

	pEvent = FLEXNET_pEventHead;

	/* Look for tail */
	while(pEvent->pNext)
		pEvent = pEvent->pNext;

	pEvent->pNext = (FLEXNET_EVENT*) malloc(sizeof FLEXNET_EVENT);
	if( pEvent->pNext == NULL )
		return;		/* Not enough memory : frame is lost (would not occur with windows)*/

	pEvent = pEvent->pNext;

	/* Copy data buffer */
	if( nDataLen > 0 )
	{
		pEvent->pszData = (char*) malloc(nDataLen + 1);
		if( pEvent->pszData == NULL )
			return;		/* Not enought memory */
		memcpy(pEvent->pszData, pszData, nDataLen);
	}
	else
		pEvent->pszData = NULL;

	pEvent->nStreamNum = nStreamNum;
	pEvent->nEventID   = nEventID;
	pEvent->nDataLen   = nDataLen;
	pEvent->pNext      = NULL;
}

static DWORD WINAPI FLEXNET_threadProc(HWND hWnd)
{
	int index;
	int nQsoNum;
	int nState;
	const INFO * info;

	for(;;)
	{
		Sleep(10);

		EnterCriticalSection(&cs);

		for(index = MIN_USERSTREAM; index <= MAX_USERSTREAM; index++)
		{
			nQsoNum = FLEXNET_nQsoNum[index];

			if( nQsoNum == 0 )
				continue;	/* Not a flexnet stream */

			/* Is that stream still connected ? */
			nState = WDXNFLX_l2state(nQsoNum);
			if( (nState & 0x1F) == 0 || (nState & 0x1F) == 1 )
			{
				FLEXNET_addEvent(index, FLEXNET_DMFRAME, NULL, 0);
				WDXNFLX_l2clri(nQsoNum);
				FLEXNET_nQsoNum[index] = 0;
				continue;
			}

			/* Is there some new frames arrived on ? */
			while( (info = WDXNFLX_l2geti(nQsoNum)) != 0 )
			{
				/* Ack the new frame and copy it to the internal buffer */
				WDXNFLX_l2iack(nQsoNum);
				FLEXNET_addEvent(index, FLEXNET_FRAME, info->text, info->len);
			}
		}

		LeaveCriticalSection(&cs);
	}
	return 0;
}

static int FLEXNET_newConnection(int nQsoNum)
{
	int   index;
	char  szCallSign[20];
	const FRAME * f;
	
	EnterCriticalSection(&cs);

	f = WDXNFLX_l2getf(nQsoNum);
	WDXNFLX_strcall((unsigned char*) szCallSign, f->dest);

	/* Maybe a link reset */
	for(index = MIN_USERSTREAM; index <= MAX_USERSTREAM; index++)
	{
		if( nQsoNum == FLEXNET_nQsoNum[index] )
		{
			strcpy(AGW_qso[index].szCallSign, szCallSign);
			AGW_qso[index].nPort  = f->kanal;
			WDXNFLX_l2sabmresp(nQsoNum);
			LeaveCriticalSection(&cs);
			return TRUE;
		}
	}

	/* Look for a free stream in dxnet */
	for(index = MIN_USERSTREAM; index <= MAX_USERSTREAM; index++)
	{
		if( AGW_qso[index].szCallSign[0] == '\0' )
			break;
	}
	
	/* No free stream available : send DM frame */
	if( index > MAX_USERSTREAM )
	{
		WDXNFLX_l2sabmdm(nQsoNum);
		LeaveCriticalSection(&cs);
		return TRUE;
	}

	/* Accept connection */
	WDXNFLX_l2sabmresp(nQsoNum);
	FLEXNET_nQsoNum[index] = nQsoNum;
	AGW_qso[index].nFlags = AGW_NEW_INCOMING_CONNECTION | AGW_FLEXNET;
	AGW_qso[index].nPort  = f->kanal;
	strcpy(AGW_qso[index].szCallSign, szCallSign);

	LeaveCriticalSection(&cs);

	return TRUE;
}

int FLEXNET_init(void)
{
	DWORD dwThreadID;

	if( gLibWdxnFlx == NULL )
	{
		if( (gLibWdxnFlx = LoadLibrary("WDXNFLX.DLL")) == NULL )
		{
			MessageBox(NULL, "Can't access the WDXNFLX.DLL file", "Error in DxNet", MB_OK);
			return FALSE;
		}
		WDXNFLX_axinit     = (WDXNFLX_AXINIT)     GetProcAddress(gLibWdxnFlx, "WDXNFLX_axinit");
		WDXNFLX_l2getsabm  = (WDXNFLX_L2GETSABM)  GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2getsabm");
		WDXNFLX_l2state    = (WDXNFLX_L2STATE)    GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2state");
		WDXNFLX_l2clri     = (WDXNFLX_L2CLRI)     GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2clri");
		WDXNFLX_l2iack     = (WDXNFLX_L2IACK)     GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2iack");
		WDXNFLX_l2sendchar = (WDXNFLX_L2SENDCHAR) GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2sendchar");
		WDXNFLX_l2pack     = (WDXNFLX_L2PACK)     GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2pack");
		WDXNFLX_l2getf     = (WDXNFLX_L2GETF)     GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2getf");
		WDXNFLX_l2ialloc   = (WDXNFLX_L2IALLOC)   GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2ialloc");
		WDXNFLX_l2unack    = (WDXNFLX_L2UNACK)    GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2unack");
		WDXNFLX_l2connect  = (WDXNFLX_L2CONNECT)  GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2connect");
		WDXNFLX_l2geti     = (WDXNFLX_L2GETI)     GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2geti");
		WDXNFLX_strcall    = (WDXNFLX_STRCALL)    GetProcAddress(gLibWdxnFlx, "WDXNFLX_strcall");
		WDXNFLX_l2stopqso  = (WDXNFLX_L2STOPQSO)  GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2stopqso");	
		WDXNFLX_l2sabmresp = (WDXNFLX_L2SABMRESP) GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2sabmresp");
		WDXNFLX_l2sabmdm   = (WDXNFLX_L2SABMDM)   GetProcAddress(gLibWdxnFlx, "WDXNFLX_l2sabmdm");	
	}
	
	if( ! WDXNFLX_axinit() )
		return FALSE;

	if( FLEXNET_initialized )
		return FALSE;	/* Flexnet already initialized */

	memset(&FLEXNET_nQsoNum, 0, sizeof FLEXNET_nQsoNum);
	FLEXNET_initialized = TRUE;

	FLEXNET_pEventHead = (FLEXNET_EVENT*) malloc(sizeof FLEXNET_EVENT);
	if( FLEXNET_pEventHead == NULL )
	{
		LOG_sysMessage(LOG_INTERNALERROR, "*** Fatal error FLEXNET_init : unable to allocate memory");		
		TOOLS_exitDxNet(10, "Not enough memory");
	}
	FLEXNET_pEventHead->pNext = NULL;

	/* Create thread */
	InitializeCriticalSection(&cs);
	CreateThread(NULL, 0, FLEXNET_threadProc, NULL , NULL, &dwThreadID);

	return TRUE;
}

/* Set my call */
void FLEXNET_setMyCall(char* pszCallsign)
{
	// pszCallsign assumed to be < 10 characters (see SWITCH_setMyCall)
	strcpy(FLEXNET_myCall, pszCallsign);
}

/* --------------------------------------------------------------------------------------------- 
   ------ Check events, like new connections, frames, ... --------------------------------------
   --------------------------------------------------------------------------------------------- 
   That function return TRUE untill an event occured */
int FLEXNET_checkEvents(void)
{
	int  nQsoNum;
	FLEXNET_EVENT* pEvent;

	if( ! FLEXNET_initialized )
		return FALSE;	/* Nothing to do there */

	EnterCriticalSection(&cs);

	/* Check for new connection */
	if( (nQsoNum = WDXNFLX_l2getsabm((unsigned char*) FLEXNET_myCall)) != 0 )
	{
		LeaveCriticalSection(&cs);
		return FLEXNET_newConnection(nQsoNum);
	}

	/* Check for incoming frames */
	pEvent = FLEXNET_pEventHead->pNext;
	while(pEvent)
	{
		FLEXNET_EVENT* pNextEvent;

		switch(pEvent->nEventID)
		{
		case FLEXNET_FRAME :
			BUFFERS_addBuff_sized(pEvent->nStreamNum, pEvent->pszData, IN, pEvent->nDataLen);
			break;

		case FLEXNET_DMFRAME :
			AGW_qso[pEvent->nStreamNum].nFlags |= AGW_NEW_INCOMING_DISCONNECT;
			break;
		}
		
		/* Delete allocated buffers */
		pNextEvent = pEvent->pNext;
		free(pEvent->pszData);
		free(pEvent);
		pEvent = pNextEvent;
	}
	FLEXNET_pEventHead->pNext = NULL;

	LeaveCriticalSection(&cs);

	return FALSE;
}

/* ------------------------------------------------------------------------------------------- 
   ------ Send a frame to flexnet ------------------------------------------------------------
   -------------------------------------------------------------------------------------------*/
void FLEXNET_sendFrame(int StreamNum, char* pszString, int nSize)
{
	if( WDXNFLX_l2ialloc(FLEXNET_nQsoNum[StreamNum], nSize) )
    {
		for(int index = 0; index < nSize; index++)
			WDXNFLX_l2sendchar(FLEXNET_nQsoNum[StreamNum], *pszString++);
		WDXNFLX_l2pack(FLEXNET_nQsoNum[StreamNum]);
    }
}

/* ------------------------------------------------------------------------------------------- 
   ------ Get unacked frames -----------------------------------------------------------------
   -------------------------------------------------------------------------------------------*/
int FLEXNET_getOutStandingFrameOnStream(int nStreamNum)
{
	int unacked;

	EnterCriticalSection(&cs);
	if( ! WDXNFLX_l2ialloc(FLEXNET_nQsoNum[nStreamNum], 256) )
		unacked = 255;
	else
		unacked = WDXNFLX_l2unack(FLEXNET_nQsoNum[nStreamNum]);
	LeaveCriticalSection(&cs);

	return unacked;
}

/* ------------------------------------------------------------------------------------------- 
   ------ Disconnect stream ------------------------------------------------------------------ 
   -------------------------------------------------------------------------------------------*/
void FLEXNET_disconnect(int StreamNum)
{
	EnterCriticalSection(&cs);
	WDXNFLX_l2stopqso(FLEXNET_nQsoNum[StreamNum]);
	LeaveCriticalSection(&cs);

	AGW_qso[StreamNum].nFlags = AGW_DISCONNECTED;
}

/* ------------------------------------------------------------------------------------------- 
   ------ Try to establish a connection ------------------------------------------------------ 
   -------------------------------------------------------------------------------------------*/
void FLEXNET_connect(int StreamNum, char* pszToString)
{
	int  nQsoNum;
	char* ptr;

	if( StreamNum < 1 || StreamNum > 64 )
		return;

	strupr(pszToString);
	TOOLS_removeNR(pszToString);

	EnterCriticalSection(&cs);	

	if( (nQsoNum = WDXNFLX_l2connect((byte*) FLEXNET_myCall, (byte*) pszToString)) == 0 )
	{
		LeaveCriticalSection(&cs);
		return;
	}
	FLEXNET_nQsoNum[StreamNum] = nQsoNum;

	LeaveCriticalSection(&cs);

	if( (ptr = strchr(pszToString, ' ')) != NULL )
		*ptr = '\0';
	TOOLS_maxLength(pszToString, 9);

	strcpy(AGW_qso[StreamNum].szCallSign, pszToString);
	AGW_qso[StreamNum].nPort  = 5;
	AGW_qso[StreamNum].nFlags = AGW_TRY_TO_CONNECT | AGW_FLEXNET;
}

/* ------------------------------------------------------------------------------------------- 
   ------ Get channel number -----------------------------------------------------------------
   -------------------------------------------------------------------------------------------*/
int FLEXNET_getPortNum(int StreamNum)
{
	const FRAME * f;

	EnterCriticalSection(&cs);
	f = WDXNFLX_l2getf(FLEXNET_nQsoNum[StreamNum]);
	LeaveCriticalSection(&cs);

	return f->kanal;
}
